
import typing

import comma.classes.file
import comma.classes.row
import comma.classes.table
import comma.helpers
import comma.typing


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "load"
]


def load(
    source: comma.typing.SourceType,
    encoding: str = None,
    delimiters: typing.Optional[typing.Iterable[str]] = None,
):
    """

    """

    # Use the helper method to open the data, parse it and return
    # a CommaInfo typed dictionary.

    csv_comma_info = comma.helpers.open_csv(
        source=source,
        encoding=encoding,
        delimiters=delimiters,
    )

    if csv_comma_info is None:
        return

    csv_rows_raw = csv_comma_info["rows"]
    csv_header = csv_comma_info["header"]

    # create CommaFile object
    parent_comma_file = comma.classes.file.CommaFile(
        header=csv_header,
        params=csv_comma_info["params"],
    )

    def _make_comma_row(csv_row_data):
        return comma.classes.row.CommaRow(
            csv_row_data,
            parent=parent_comma_file,
        )

    csv_comma_rows = list(map(_make_comma_row, csv_rows_raw))

    csv_comma_table = comma.classes.table.CommaTable(
        csv_comma_rows,
        parent=parent_comma_file
    )

    return csv_comma_table





import collections

import comma.exceptions


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CommaFieldSlice",
    "CommaRowSlice"
]


class CommaFieldSlice(list, collections.UserList):
    """
    Contains a column of a CSV file.
    """

    # parent CSV file
    _parent = None

    # field name
    _field_name = None

    def __init__(self, initlist=None, parent=None, field_name=None, *args, **kwargs):

        self._parent = parent
        self._field_name = field_name
        self._field_index = self._parent.header.index(self._field_name)

        ret = super().__init__(initlist, *args, **kwargs)
        return ret

    def __iter__(self):
        for row in super().__iter__():
            yield row[self._field_index]

    def __getitem__(self, key):
        if type(key) is int:
            ret_val = super().__getitem__(key)
            return ret_val[self._field_index]

        elif type(key) is slice:
            ret_slice = super().__getitem__(key)

            return CommaFieldSlice(
                initlist=ret_slice,
                parent=self._parent,
                field_name=self._field_name)

    def __setitem__(self, key, value):
        if type(key) is int:
            row = super().__getitem__(key)
            row[self._field_index] = value

        elif type(key) is slice:
            ret_slice = super().__getitem__(key)

            if len(ret_slice) != len(value):
                raise comma.exceptions.CommaBatchException(
                    ("trying to update a slice ({}) with a "
                     "range of the wrong size ({})").format(
                        len(ret_slice),
                        len(value),
                    ))

            for i in range(len(ret_slice)):
                ret_slice[i][self._field_index] = value[i]

            return ret_slice


class CommaRowSlice(collections.UserList, list, collections.UserDict):
    """
    Contains a column of a CSV file.
    """

    # parent CSV file
    _parent = None

    # header slice
    _header_slice = None

    def __init__(self, initlist=None, parent=None, header_slice=None, *args, **kwargs):

        self._parent = parent
        self._header_slice = header_slice

        ret = super().__init__(initlist, *args, **kwargs)
        return ret

    def __getitem__(self, key):
        ##print("CommaRowSlice.__getitem__", hex(id(self)), key, type(key))

        if type(key) is int:
            ret_val = super().__getitem__(key)
            return ret_val[self._field_index]

        elif type(key) is slice:
            ret_slice = super().__getitem__(key)

            return CommaFieldSlice(
                initlist=ret_slice,
                parent=self._parent,
                field_name=self._field_name)

    def __setitem__(self, key, value):
        ##print("CommaRowSlice.__setitem__", hex(id(self)), key, type(key), "<==", value, type(value))

        if type(key) is int:
            row = super().__getitem__(key)
            row[self._field_index] = value

        elif type(key) is slice:
            ret_slice = super().__getitem__(key)

            if len(ret_slice) != len(value):
                raise comma.exceptions.CommaBatchException(
                    ("trying to update a slice ({}) with a "
                     "range of the wrong size ({})").format(
                        len(ret_slice),
                        len(value),
                    ))

            for i in range(len(ret_slice)):
                ret_slice[i][self._field_index] = value[i]

            return ret_slice




import typing

import comma.exceptions
import comma.typing


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CommaFile"
]


class CommaFile(object):
    """
    """

    # Internal instance variable containing header
    _header = None

    # Internal instance variable with params
    _params = None

    # "Primary key" through which to access the records
    _primary_key = None

    def __init__(
            self,
            header=None,
            params: comma.typing.CommaInfoParams = None,
            primary_key: typing.Optional[str] = None
    ):
        self._header = header
        self._params = params
        self._primary_key = primary_key

    @property
    def header(self):
        """

        """
        return self._header

    @header.setter
    def header(self, value):
        raise NotImplementedError("not yet implemented")

    @property
    def primary_key(self):
        """

        """
        return self._primary_key

    @primary_key.setter
    def primary_key(self, value):

        # shortcuts to un-setting the primary key
        if value is None or value == "" or value == False:
            del self.primary_key
            return

        # from this point on, consider we are setting the primary key

        # first check whether there are headers
        if self.header is None:
            raise comma.exceptions.CommaNoHeaderException(
                "cannot set the primary key of a `CommaFile` "
                "that does not have a header"
            )

        # next check if proposed header belongs to headers
        if value not in self.header:
            # Try to get a string representation of headers for diagnostic
            # purposes for user; yes, the exception is too broad because
            # we don't really care why the headers couldn't be converted
            # to string

            # Thanks @pylover!
            # See: https://gist.github.com/pylover/7870c235867cf22817ac5b096defb768

            # noinspection PyBroadException
            try:
                header_string = self.header.__repr__()
            except Exception:  # pylint: disable=too-general-exception
                header_string = ""

            raise comma.exceptions.CommaKeyError(
                "the requested primary key (" +
                value +
                ") is not one of the headers: " +
                header_string
            )

        # seems we have validated the primary key and so setting it
        self._primary_key = value

    @primary_key.deleter
    def primary_key(self):
        self._primary_key = None





__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CommaException",
    "CommaOrphanException",
    "CommaOrphanRowException",
    "CommaOrphanTableException",
    "CommaNoHeaderException",
    "CommaKeyError",
    "CommaBatchException",
]


class CommaException(Exception):
    """
    The base exception for the `comma` package.
    """
    pass


class CommaOrphanException(CommaException):
    """
    An internal reference has been broken.
    """
    pass


class CommaOrphanRowException(CommaOrphanException):
    """
    A row required access to information from its parent CSV file but was
    unable to.
    """
    pass


class CommaOrphanTableException(CommaOrphanException):
    """
    A table required access to information from its parent CSV file but was
    unable to.
    """
    pass


class CommaNoHeaderException(CommaException, KeyError):
    """
    A header was expected (or necessary to an operation) but was not found.
    """
    pass


class CommaKeyError(CommaException, KeyError):
    """
    The requested key is not part of the header of this file.
    """
    pass


class CommaBatchException(CommaException):
    """
    A batch update was not possible, because invalid.
    """
    pass



import codecs
import collections
import csv
import io
import os
import pathlib
import urllib, urllib.parse
import zipfile

import comma.helpers


__author__ = ("Jérémie Lumbroso <lumbroso@cs.princeton.edu>")

__all__ = [
    "detect_csv_type",
    "detect_encoding",
    "is_binary_string",
]


DEFAULT_DELIMITER = u","


# Better CSV dialect detection, thanks to clevercsv

# define a helper based on the standard CSV package
def _default_detect_csv_type(sample, delimiters=None):
    """
    Returns a dictionary containing meta-data on a CSV file, such as
    the format "dialect", whether the file is likely to have a header
    and the kind of line terminator that has been detected. This
    version of the helper method is based on Python internal's `csv`
    module.
    """

    sniffer = csv.Sniffer()
    truncated_sample = sample[:comma.helpers.MAX_SAMPLE_CHUNKSIZE]

    # better line terminator detection
    line_terminator = comma.helpers.detect_line_terminator(truncated_sample)

    try:
        dialect = sniffer.sniff(sample=truncated_sample, delimiters=delimiters)
        dialect.lineterminator = line_terminator

        has_header = sniffer.has_header(sample=truncated_sample)
    except csv.Error:
        dialect = None
        has_header = False

    return {
        "dialect": dialect,
        "simple_dialect": None,
        "has_header": has_header,
        "line_terminator": line_terminator,
    }


detect_csv_type = _default_detect_csv_type

try:
    import clevercsv
    
    def detect_csv_type(sample, delimiters=None):
        """
        Returns a dictionary containing meta-data on a CSV file, such as
        the format "dialect", whether the file is likely to have a header
        and the kind of line terminator that has been detected. This
        version of the helper method is based on the excellent/essential
        Python package `clevercsv` by Gertjan van den Burg (@GjjvdBurg),
        see https://github.com/alan-turing-institute/CleverCSV.
        """

        sniffer = clevercsv.Sniffer()
        truncated_sample = sample[:comma.helpers.MAX_SAMPLE_CHUNKSIZE]
        simple_dialect = sniffer.detect(sample=truncated_sample, delimiters=delimiters)
        line_terminator = comma.helpers.detect_line_terminator(truncated_sample)

        dialect = simple_dialect.to_csv_dialect()
        dialect.lineterminator = line_terminator

        has_header = False
        try:
            has_header = sniffer.has_header(sample=truncated_sample)
        except StopIteration:
            # can happen with empty data
            pass

        # also a fix for empty streams
        if dialect.delimiter is None or dialect.delimiter == "":
            dialect.delimiter = DEFAULT_DELIMITER

        return {
            "dialect": dialect,
            "simple_dialect": simple_dialect,
            "has_header": has_header,
            "line_terminator": line_terminator,
        }

except ImportError:  # pragma: no cover
    clevercsv = None


# Better detection of binary data (i.e., zipped files), thanks to binaryornot

# In case the package is not available: Define our own helper method
# Based on file(1), see https://stackoverflow.com/a/7392391/408734
TEXT_CHARS = bytearray({7,8,9,10,12,13,27} | set(range(0x20, 0x100)) - {0x7f})


def _is_binary_string_internal(bytestring):
    return bool(bytestring.translate(None, TEXT_CHARS))


_is_binary_string = _is_binary_string_internal

try:
    # See https://github.com/audreyr/binaryornot/
    import binaryornot
    import binaryornot.helpers
    
    # Alias to helper method
    _is_binary_string = binaryornot.helpers.is_binary_string

except ImportError:  # pragma: no cover
    binaryornot = None


def is_binary_string(bytestring, truncate=True):
    """
    Detect, using heuristics, whether a string of bytes is text or binary data.
    If available, this will use the `binaryornot` lightweight package.
    """
    
    if bytestring is None:
        return False
    
    bytestring_length = -1
    try:
        bytestring_length = len(bytestring)
        
    except TypeError:
        return False
    
    if truncate and bytestring_length > comma.helpers.MAX_SAMPLE_CHUNKSIZE:
        return _is_binary_string(bytestring[:comma.helpers.MAX_SAMPLE_CHUNKSIZE])
    
    return _is_binary_string(bytestring)


# Better encoding detection

# Borrowed from requests.util, helper method to detect encoding of JSON
# outputs returned by the HTTP requests.
#
# See https://github.com/psf/requests/blob/3e7d0a873f838e0001f7ac69b1987147128a7b5f/requests/utils.py#L856-L891


# Null bytes; no need to recreate these on each call to guess_json_utf

_null = '\x00'.encode('ascii')  # encoding to ASCII for Python 3
_null2 = _null * 2
_null3 = _null * 3


def _detect_encoding_by_bom(sample, default=None):
    """
    :rtype: str
    """

    # JSON always starts with two ASCII characters, so detection is as
    # easy as counting the nulls and from their location and count
    # determine the encoding. Also detect a BOM, if present.

    sample = sample[:4]

    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return 'utf-32'     # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return 'utf-8-sig'  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return 'utf-16'     # BOM included

    nullcount = sample.count(_null)

    if nullcount == 0:
        return default

    if nullcount == 2:
        if sample[::2] == _null2:   # 1st and 3rd are null
            return 'utf-16-be'
        if sample[1::2] == _null2:  # 2nd and 4th are null
            return 'utf-16-le'
        # Did not detect 2 valid UTF-16 ascii-range characters

    if nullcount == 3:
        if sample[:3] == _null3:
            return 'utf-32-be'
        if sample[1:] == _null3:
            return 'utf-32-le'
        # Did not detect a valid UTF-32 ascii-range character

    return default


detect_encoding = _detect_encoding_by_bom

# If chardet is available, use it as a second round of guessing
# if the previous method was unsuccessful

try:
    import chardet
    
    def detect_encoding(sample, default="utf-8"):
        # First try a fool-proof deterministic method
        encoding = _detect_encoding_by_bom(sample)
        if encoding is not None:
            return encoding
        
        # If that doesn't work, try a heuristic
        result = chardet.detect(sample)
        if result is not None and result.get("encoding") is not None:
            return result.get("encoding")
        
        return default
    
except ImportError:  # pragma: no cover
    chardet = None






import copy
import typing


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CloneableCollection"
]


class CloneableCollection(object):
    """
    An abstract collection (containing a `data` attribute, for instance, from
    `collections.UserList` or `collections.UserDict`) that can be cloned with
    new data.
    """

    # noinspection PyMethodMayBeStatic
    def _validate_newdata(self, newdata: typing.Any = None) -> bool:
        """
        Returns `True` if the provided `newdata` is acceptable as the internal
        state of the collection. This protected helper method is called by
        `CloneableCollection.clone()` to ensure the new data provided for a
        cloned object will not lead the object to be an invalid internal state.
        """

        if newdata is None:
            return True

        return True

    def clone(self, newdata: typing.Any = None, **kwargs):
        """
        Returns a clone of the current collection, with possible different
        underlying data, as specified by `newdata`.
        """

        if not hasattr(self, "data"):
            raise Exception(
                "for a class to be a `ClonableCollection`, it must at"
                "at least have a `data` attribute"
            )

        if "data" in kwargs:
            raise ValueError(
                "use the `newdata` keyword too override existing data (that way "
                "the data is properly validated, unlike for all other overrides)"
            )

        # validate new data
        if not self._validate_newdata(newdata=newdata):
            raise ValueError("`newdata` failed internal validation")

        # create new class instance
        inst = self.__class__.__new__(self.__class__)

        # Copy all attributes except obj.data (handled separately)
        for key, value in self.__dict__.items():
            # kwargs can override existing attributes
            if key in kwargs:
                inst.__dict__ [key] = kwargs.get(key)

            # this is handled separately
            if key != "data":
                inst.__dict__[key] = value

        # Create a copy of the data and avoid triggering descriptors
        inst.__dict__["data"] = newdata or copy.deepcopy(self.__dict__["data"])

        return inst




from comma.classes.file import CommaFile
from comma.classes.table import CommaTable
from comma.classes.row import CommaRow
from comma.classes.slices import CommaFieldSlice, CommaRowSlice


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CommaFile",
    "CommaTable",
    "CommaRow",
    "CommaFieldSlice",
    "CommaRowSlice",
]



import collections

import comma.exceptions
import comma.helpers


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CommaRow"
]


class CommaRow(collections.UserList, list, collections.UserDict):
    """
    Contains a single row of a CSV file; the row contains only data
    stored in the row, and a pointer to a parent file structure (which
    stores all the extraneous information, such as dialect and header).
    """

    # parent CSV file
    _parent = None

    # (optionally) sequence of slicing operations
    _slice_list = None

    # (optionally) original row reference
    _original = None

    def __init__(self, initlist=None, parent=None, slice_list=None, original=None, *args, **kwargs):
        # initialize internal attributes
        self._parent = parent
        self._slice_list = slice_list if slice_list is not None else list()
        self._original = self if original is None else original

        # call base constructor for lists
        ret = super().__init__(initlist, *args, **kwargs)

    @property
    def header(self):
        """
        """
        if self._parent is None:
            raise comma.exceptions.CommaOrphanRowException(
                "row not linked to parent structure")

        if self._parent.header is None:
            raise comma.exceptions.CommaNoHeaderException(
                "CSV file does not appear to have a header and "
                "none was provided; key-based access not possible")

        return comma.helpers.multislice_sequence(
            sequence=self._parent.header,
            slice_list=self._slice_list)

    def keys(self):
        return self.header

    def values(self):
        return dict(self).values()

    def items(self):
        return dict(self).items()

    def __key_to_column_id(self, key):
        """

        """

        # CASE 1: a regular list index
        if type(key) is int:
            return key

        # CASE 2: a slice
        if type(key) is slice:
            slice_start = key.start
            slice_stop = key.stop

            if slice_start is not None:
                slice_start = self.__key_to_column_id(slice_start)

            if slice_stop is not None:
                slice_stop = self.__key_to_column_id(slice_stop)

            return slice(slice_start, slice_stop, key.step)

        # CASE 3: a dictionary key
        header = self.header

        if not key in header:
            raise comma.exceptions.CommaKeyError(
                "{key} is not in header: {header}".format(
                    key=key,
                    header=header))

        key_index = header.index(key)

        return key_index

    def __setitem__(self, key, value):
        ##print("CommaRow.__setitem__", hex(id(self)), key, type(key), "<==", value, type(value))

        # FIXME: figure out how to make slices have headers
        key_index = self.__key_to_column_id(key)
        if type(key) is str and self._original != self:
            ##print(type(key) is str and self._original != self)
            return self._original.__setitem__(key, value)
        ret = super().__setitem__(key_index, value)
        return ret

    def __getitem__(self, key):
        ##print("CommaRow.__getitem__", hex(id(self)), key, type(key))

        # FIXME: figure out how to make slices have headers
        key_index = self.__key_to_column_id(key)
        ret = super().__getitem__(key_index)

        if type(key) is slice:
            ret._parent = self._parent
            ret._slice_list.append(key)
            ret._original = self._original

        return ret

    def __repr__(self):
        # display as list or row
        try:
            self.keys()
        except comma.exceptions.CommaException:
            return super().__repr__()
        return dict(self).__repr__()




import csv
import io
import os
import typing
import urllib, urllib.parse
import zipfile

try:
    import requests
except ImportError:  # pragma: no cover
    requests = None

import comma.extras
import comma.typing


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "MAX_SAMPLE_CHUNKSIZE",
    "URI_SCHEME_LOCAL",
    "URI_SCHEMES_ACCEPTED",
    "LINE_TERMINATORS",
    "LINE_TERMINATOR_DEFAULT",

    "is_local",
    "is_url",
    "detect_line_terminator",
    "open_stream",

    "multislice_sequence",
]


MAX_SAMPLE_CHUNKSIZE = 10000

URI_SCHEME_LOCAL = "file"

URI_SCHEMES_ACCEPTED = ["http", "https"]

LINE_TERMINATORS = ["\r\n", "\r", "\n"]

LINE_TERMINATOR_DEFAULT = "\n"


def is_anystr(obj: typing.Union[typing.Any, typing.AnyStr]) -> bool:
    """
    Returns `True` if the `obj` object is of type `typing.AnyStr`.
    """
    return (
            obj is not None and
            (isinstance(obj, str) or
             isinstance(obj, bytes)))


def is_local(location: typing.AnyStr) -> typing.Optional[str]:
    """
    Detects whether a string location is a local file path.
    """
    
    # Eliminate obvious non-paths
    if location is None or location == "":
        return
    
    # Try to parse the location using urlparse (to handle file://)
    parsed_location = None
    
    try:        
        parsed_location = urllib.parse.urlparse(location)
        
        # parsed_location = ParseResult(
        #   scheme=..., netloc=..., path=...,
        #   params=..., query=..., fragment=...)

    except TypeError:
        parsed_location = None
    except AttributeError:
        parsed_location = None

    # May not be a string; regular paths should be parsed without
    # trouble
    if parsed_location is None:
        return
        
    path = None
    
    if parsed_location.scheme in [URI_SCHEME_LOCAL, ""]:
        
        parsed_path = parsed_location.path
        
        if parsed_location.netloc != "":
            parsed_path = parsed_location.netloc + parsed_path
        
        if os.path.exists(parsed_path):
            path = parsed_path
        
        if os.path.exists(os.path.expanduser(parsed_path)):
            path = os.path.expanduser(parsed_path)
    
    if os.path.exists(location):
        path = location
    
    elif os.path.exists(os.path.expanduser(location)):
        path = os.path.expanduser(location)
    
    if path is not None:
        path = os.path.abspath(path)
        return path
    
    return


def is_url(location: str, no_request: bool = False) -> bool:
    """
    Detects whether a string location is a URL; may make a test HEAD request
    if the location is likely to be an actual URL (this behavior can be
    deactivated by setting `no_request` to `True`).
    """
    
    # Eliminate obvious non-URL
    if location is None or location == "":
        return False
    
    # Try to parse the URL using urlparse
    parsed_location = None
    
    try:        
        parsed_location = urllib.parse.urlparse(location)
        
        # parsed_location = ParseResult(
        #   scheme=..., netloc=..., path=...,
        #   params=..., query=..., fragment=...)

    except AttributeError:
        # May not be a string
        parsed_location = None
    
    except Exception as exc:
        # Unexpected error
        parsed_location = None
    
    if parsed_location is None:
        return False
    
    # Check parsed location
    if parsed_location.scheme == "" or parsed_location.netloc == "":
        return False
    
    # This is an actual file
    if parsed_location.scheme == URI_SCHEME_LOCAL:
        return False
    
    # If we cannot make an actual HEAD request, then this is 
    if no_request or requests is None:
        return parsed_location.scheme in URI_SCHEMES_ACCEPTED
    
    response = None
    
    # Try to make a HEAD request on the location to see if it is successful
    try:
        response = requests.head(location, allow_redirects=True, timeout=10)
    
    except requests.exceptions.InvalidSchema:
        # Not a supported scheme
        return False
    
    except requests.exceptions.ConnectionError:
        # Not able to connect
        return False
    
    if response is None:
        return False
    
    return response.ok


def detect_line_terminator(
        sample: typing.Optional[typing.AnyStr],
        default: typing.Optional[typing.AnyStr] = None
) -> str:
    """
    Detects the most likely line terminator (from `\r`, `\n`, `\r\n`), given
    a sample string, by counting the occurrences of each pattern and finding
    the longest and most frequent.
    """

    # update default
    if default is None:
        default = LINE_TERMINATOR_DEFAULT

    if sample is None or not hasattr(sample, "count"):
        return default

    # check to see if can obtain valid counts for all line terminators
    for lt in LINE_TERMINATORS:
        try:
            val = sample.count(lt)
            if type(val) is not int:
                return default
        except TypeError:
            return default

    # the sorting of options is made taking into account both
    # the number of occurrences of a pattern, and the length of
    # the pattern (this is so when "\r\n" occurs, it also boosts
    # the count of "\r" and "\n", so we must also look at the
    # LONGEST pattern with the best number of occurrences)

    ranked_options = sorted(
        zip(
            map(sample.count, LINE_TERMINATORS),  # counts
            map(len, LINE_TERMINATORS),           # length
            LINE_TERMINATORS),                    # line terminators
        reverse=True)

    best_option = ranked_options[0]

    # if the best option has been counted 0 times, means no line terminators
    # were found
    if best_option[0] == 0:
        return default

    # otherwise return the value of the best option
    return best_option[2]


def open_stream(
    source: comma.typing.SourceType,
    encoding: str = None,
    no_request: bool = False,
) -> typing.Optional[typing.TextIO]:
    """
    Returns a seekable stream for text data that is properly decoded
    and ready to be read: The `source` can be actual data, a local file
    path, or a URL; it is possible to provide a stream that is compressed
    using ZIP. (This method will store all the data in memory.)
    """
    
    if source is None:
        return

    # local variable to keep track of the (most accurate for the user)
    # caption of the source
    internal_name = None
    
    # is this a STRING?
    if type(source) is str:
        source = typing.cast(typing.AnyStr, source)
        
        # multiline?
        if "\n" in source or "\r" in source:
            newline = comma.helpers.detect_line_terminator(
                sample=source,
                default="\n")

            # change to a standard newline
            source = source.replace(newline, "\n")

            return io.StringIO(initial_value=source, newline="\n")
        
        internal_name = source
        
        # is this a FILE?
        local_path = is_local(location=source)
        if local_path is not None:
            source = open(local_path, mode="rb")
        
        # is this a URL?
        elif not no_request and is_url(location=source):
            
            response = requests.get(url=source, allow_redirects=True)
            
            if not response.ok:
                return None
            
            if encoding is None:
                encoding = response.encoding
            
            if encoding is not None:
                source = io.StringIO(response.text)
            else:
                source = io.BytesIO(response.content)
        
        else:
            return None

    # is this BYTES?
    if type(source) is bytes:
        source = typing.cast(bytes, source)
        source = io.BytesIO(source)

    # is this a STREAM?
    if hasattr(source, "seekable"):
        
        # is it not seekable? if so, make it seekable
        if not source.seekable():
            
            # if not, read in all the data
            data = source.read()
            
            if type(data) is str:
                source = io.StringIO(data)
            
            elif type(data) is bytes:
                source = io.BytesIO(data)
        
        # is it compressed? if so, unzip it
        if zipfile.is_zipfile(source):
            zipsource = zipfile.ZipFile(source, mode="r")
            
            names = zipsource.namelist()
            
            count_total = 0
            count_csv = 0
            
            csv_filename = None
            
            for name in names:
                count_total += 1
                if os.path.splitext(name)[1].lower() == ".csv":
                    count_csv += 1
                    csv_filename = name
            
            if count_total == 1:
                # if only one file, we don't care if it is a CSV (we assume)
                data = zipsource.read(name=names[0])
                source = io.BytesIO(data)
            
            elif count_total > 1 and count_csv == 1:
                # if exactly one CSV, we know what to do
                data = zipsource.read(name=csv_filename)
                source = io.BytesIO(data)
            
            elif count_total == 0:
                raise ValueError(
                    "it seems the provided source is ZIP compressed; but "
                    "there are unknown issues unzipping it (or the archive "
                    "is empty)")
                
            else:
                # other situations are unclear
                raise ValueError(
                    "provided ZIP source is ambiguous, "
                    "contains multiple files: {}".format(names))
    
    # if at this point, has not been converted to stream, error
    if not hasattr(source, "seekable"):
        return None
    
    # look at a sample and analyze
    source.seek(0)
    sample = source.read(MAX_SAMPLE_CHUNKSIZE)
    source.seek(0)   # fixed this bug with tests! :-)
    
    # detect encoding if bytestring
    if type(sample) is bytes:
        if encoding is None:
            encoding = comma.extras.detect_encoding(sample)
        source = io.TextIOWrapper(source, encoding=encoding)
    
    # try to add useful metadata
    if internal_name is not None:
        try:
            source.buffer.name = internal_name
        except AttributeError:
            pass
    
    return source


def open_csv(
        source: comma.typing.SourceType,
        encoding: str = None,
        delimiters: typing.Optional[typing.Iterable[str]] = None,
        no_request: bool = False,
) -> comma.typing.CommaInfo:
    """
    Returns a `CommaInfo` typed dictionary containing the data and
    metadata related to a CSV file. The `source` can be actual data,
    a local file path, or a URL; it is possible to provide a stream
    that is compressed using ZIP.

    The `source` is opened using the `comma.helpers.open_stream()`
    helper method. The metadata data is detected using internal
    helpers and either the `csv` or `clevercsv` dialect sniffers.
    """

    stream = comma.helpers.open_stream(
        source=source,
        encoding=encoding,
        no_request=no_request,
    )
    if stream is None:
        return

    # close at end if a stream was opened by this method (but not if
    # a stream was provided to this method)
    try:
        close_at_end = (
                not hasattr(source, "seekable") or
                source.buffer.name != stream.buffer.name)
    except AttributeError:
        close_at_end = True

    # streams returned by open_stream should be seekable
    # FIXME: handle non-seekable streams
    assert(hasattr(stream, "seekable"))

    # get a sample and analyze
    stream.seek(0)
    csv_sample = stream.read(comma.helpers.MAX_SAMPLE_CHUNKSIZE)
    stream.seek(0)

    csv_params = comma.extras.detect_csv_type(
        sample=csv_sample,
        delimiters=delimiters)

    reader = csv.reader(stream, dialect=csv_params["dialect"])
    csv_rows = [row for row in reader]

    # close if necessary
    if close_at_end:
        stream.close()

    data = {
        "rows": csv_rows,
        "params": csv_params,
        "sample": csv_sample,
        "header": None,
    }

    # isolate the headers if they exist
    if data["params"].get("has_header", False):

        if len(csv_rows) == 0:
            data["params"]["has_header"] = False
            data["column_count"] = 0

        else:
            data["header"] = csv_rows[0]
            data["rows"] = csv_rows[1:]
            data["column_count"] = len(data["header"])

    if len(csv_rows) > 0 and "column_count" not in data:
        data["column_count"] = max(map(len, csv_rows))

    # store the source location if it was a string
    if comma.helpers.is_anystr(source):
        data["source"] = source

    return data


def multislice_sequence(
        sequence: typing.Sequence[typing.Any],
        slice_list: typing.List[slice] = None,
) -> typing.Sequence[typing.Any]:
    """
    Returns the sub-sequence obtained from sequentially slicing the
    sequence `sequence` according to the series of slices in `slice_list`.
    """
    new_sequence = sequence

    if slice_list is not None:
        for sl in slice_list:
            new_sequence = new_sequence.__getitem__(sl)

    return new_sequence


# NOTE: Have not decided whether to use this or not yet
#
# def comment_stripper(stream: typing.TextIO, comment_line_chars: str = '#;'):
#     for line in stream.readlines():
#
#         stripped_line = line.strip()
#         if not stripped_line:
#             # empty lines
#             continue
#
#         # look at first character
#         if stripped_line[:1] in comment_line_chars:
#             # lines that are commented out
#             continue
#
#         # yield line
#         yield line



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['comma', 'comma.classes']

package_data = \
{'': ['*']}

extras_require = \
{'autodetect': ['chardet>=3.0.4,<4.0.0',
                'binaryornot>=0.4.4,<0.5.0',
                'clevercsv>=0.6.3,<0.7.0'],
 'net': ['requests>=2.23.0,<3.0.0']}

setup_kwargs = {
    'name': 'comma',
    'version': '0.1.2',
    'description': 'Python CSV, and delimiter-spaced files, for humans!',
    'long_description': '# Comma: A Python CSV Library for Humans\n\nThis library tries to make manipulating CSV files a great experience.\n\n## Why?\n\nAlthough Python, fortuitously, is\n"[batteries included](https://www.python.org/dev/peps/pep-0206/#batteries-included-philosophy)",\non occasion, some of the libraries end up being designed with APIs\nthat don\'t map well to what turns out to be the most common usage\npatterns. This is what happened with the various `urllib` libraries,\nincredibly powerful, but limiting users by its complexity---it was\nnot straightforward, for instance, to use cookies: One of several\nproblems that [`requests`](https://github.com/psf/requests) by\n[@ken-reitz](https://github.com/ken-reitz) addressed. Indeed,\n`requests` abstracts power beneath simplicity, smart defaults, and\ndiscoverability.\n\nFor the CSV format, we are confronted with a similar situation. While\nboth the JSON and YAML formats have packages that provide, one-command\nmeans to load content from files in those respective formats to a\nnested Python object, for the CSV format,\n[the standard library](https://www.python.org/dev/peps/pep-0305/) has\nyou use an iterator to access the data. Many details require significant\nsyntax change (for instance the difference between have lists or\ndictionaries depends on the class that is used to read the file).\n\nSince then, we also have several excellent libraries that, by providing\ngreat auto-detection (of dialect, file format, encoding, etc.) allow\nfor hiding many details from the end user.\n\nAll this to say, `comma` will try to do exactly what you want\nwhen you do:\n\n```python\nimport comma\ndata = comma.load(...)\n``` \n\n## Alternatives\n\nPython is fortunate to have a lot of very good libraries to read/write\nCSV and tabular files in general. (Some of these were discovered through\nthe excellent [Awesome Python](https://awesome-python.com/) list.)\n\n- [`clevercsv`](https://github.com/alan-turing-institute/CleverCSV): An\n  exceptional library by @GjjvdBurg, builds on statistical and empirical\n  to provide powerful and reliable CSV dialect detection. However, it\n  strives to be a drop-in replacement for the original Python `csv`\n  module, and as such does not improve on the complex syntax. This\n  library is [the culmination of serious peer-reviewed\n  research](https://arxiv.org/abs/1811.11242), and `comma` uses it\n  internally to improve auto-detection.\n  \n- [`csvkit`](https://github.com/wireservice/csvkit): This is a set of\n  command-line tools (rather than a module/package) written in Python,\n  to make it easier to manipulate CSV files. One of the highlights is\n  a tool called `csvpy <file.csv>` to open a Python shell with the CSV\n  data loaded into a Python object called `reader`, to quickly run\n  some Python logic on the data. While it is technically possible to\n  use `csvkit`\'s internals in a project, this is not documented.\n  \n- [`pandas`](https://github.com/pandas-dev/pandas): An advanced data\n  science package for Python, this certainly provides a powerful CSV\n  (and more generally, table file) reader and parser. The API of the\n  table object is very powerful, but you need to take the time to learn\n  how to use it. This library is perhaps not ideal for file manipulations.\n\n- [`pyexcel`](https://github.com/pyexcel/pyexcel): This library provides\n  access to Excel and other tabular formats, including CSV, and various\n  data sources (stream, database, file, ...). It emphasizes one common\n  format-agnostic API, that instead has the user choose the data format\n  (list, matrix, dictionary, ...).\n\n- [`tablib`](https://github.com/jazzband/tablib/): This library was\n  originally written by [Kenneth Reitz](https://github.com/ken-reitz),\n  the creator who brought [`requests`](https://github.com/psf/requests),\n  [`pipenv`](https://github.com/pypa/pipenv) and many other goodies to\n  Python---and then included in the [Jazzband](https://jazzband.co/)\n  collective. The focus of this library is on interoperating between many\n  different file formats (such as XLS, CSV, JSON, YAML, DF, etc., ...,\n  even LaTeX `booktabs`!). It seems to have a very high adoption rate\n  because it is a dependency for many Jazzband libraries. The API is\n  class-based rather than method-based. A companion library,\n  [`prettytable`](https://github.com/jazzband/prettytable) focuses on\n  pretty printing tabular data (including from a CSV file).\n\n- [`tabulator`](https://github.com/frictionlessdata/tabulator-py): This\n  library provides a single interface to manipulate extremely large\n  tabular data---and useful for files so large that they need to be\n  streamed line-by-line; the library supports a broad array of formats\n  including reading data directly from Google Spreadsheets. However\n  this power means that reading a CSV file requires several operations.\n\n### Miscellaneous\n\nAlthough not specifically a Python library, nor designed to read/write CSV\nfiles (but instead to compare them), [`daff`](https://github.com/paulfitz/daff)\nis a really cool project: It provides a `diff` of tabular data [with cell-level\nawareness](http://specs.frictionlessdata.io/tabular-diff/).\n\nAnother unrelated project is [https://getgrist.com](Grist), a spreadsheet\nPaaS, which among other useful features, allows [the use of Python within\nformulas](https://support.getgrist.com/formulas/). \n\n## Acknowledgements\n\nThanks to [@zbanks](https://github.com/zbanks) for the name of the package!\nThanks to [@rfreling](https://github.com/rfreling),\n[@adamfinkelstein](https://github.com/adamfinkelstein) for discussing ideas\nbefore I got started on this. Thanks to [@GjjvdBurg](https://github.com/GjjvdBurg)\nand collaborators for awesome, awesome contribution to text processing science\nand our Python community with `clevercsv`. \n\n## License\n\nThis project is licensed under the LGPLv3 license, with the understanding\nthat importing a Python modular is similar in spirit to dynamically linking\nagainst it.\n\n- You can use the library `comma` in any project, for any purpose, as long\n  as you provide some acknowledgement to this original project for use of\n  the library.\n\n- If you make improvements to `comma`, you are required to make those\n  changes publicly available.\n\n',
    'author': 'Jérémie Lumbroso',
    'author_email': 'lumbroso@cs.princeton.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jlumbroso/comma',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




import csv
import typing


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "SourceType",

    "DialectType",
    "SimpleDialectType",

    "CommaInfo",
    "CommaInfoParams"
]


# Our type hint for a data source:
#  - a location (URL or file path), or string data
#  - a stream (text or binary)
#  - a string of bytes

SourceType = typing.Union[typing.AnyStr, typing.IO, bytes]


# The original CSV Dialect type from Python library

DialectType = csv.Dialect


# The simplified CSV dialect description from the excellent
# https://github.com/alan-turing-institute/CleverCSV/blob/master/clevercsv/dialect.py

SimpleDialectType = typing.Any
try:
    import clevercsv
    import clevercsv.dialect
    SimpleDialectType = clevercsv.dialect.SimpleDialect
except ImportError:  # pragma: no cover
    clevercsv = None


# Type definitions for helper dictionaries

CommaInfoParams = typing.TypedDict(
    "CommaInfoParams", {
        #
        "dialect":         DialectType,

        #
        "simple_dialect":  SimpleDialectType,

        "has_header":      bool,

        "line_terminator": str,
    })

CommaInfo = typing.TypedDict(
    "CommaInfo", {
        # the parsed CSV rows
        "rows":   typing.List[typing.List[str]],

        # a raw sample of the original file
        "sample": str,

        # an identifier for the source (if not raw buffer)
        "source_name": str,

        # CSV parameters
        "header": typing.Optional[typing.List[str]],
        "params": CommaInfoParams,
    })



import collections

import comma.abstract
import comma.exceptions
import comma.classes.slices


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "CommaTable"
]


class CommaTable(collections.UserList, list, collections.UserDict, comma.abstract.CloneableCollection):
    """
    Contains a table from a CSV file.
    """

    # parent CommaFile object from which to get settings and headers
    _parent = None

    def __init__(self, initlist=None, parent=None, *args, **kwargs):
        self._parent = parent
        ret = super().__init__(initlist, *args, **kwargs)
        return ret

    def to_html(self):
        """
        Returns an HTML string representation of the table data.
        """

        # NOTE: should we use tabulator or prettytable instead of reinventing?

        # contain all the rows
        table_rows = []

        def _zip_html_tag(data, in_pattern="<td style='text-align: left;'>{}</td>", out_pattern="<tr>{}</tr>", indent=0):

            if type(indent) is not int:
                indent = 0

            linebreak = "\n" + (indent * " ")

            inner_html = linebreak.join(map(in_pattern.format, data))
            outer_html = out_pattern.format(inner_html)

            return outer_html

        if self.header is not None:
            table_rows = [_zip_html_tag(data=self.header, in_pattern="<th>{}</th>")]

        for row in self.data:
            table_rows += [_zip_html_tag(data=row)]

        table_html = "<table style='text-align: left;'>{}</table>".format("\n\n".join(table_rows))

        return table_html

    def _repr_html_(self):
        """
        Returns an HTML string representation of the table data, this is a magic
        helper method for IPython/Jupyter notebooks.
        """
        return self.to_html()

    @property
    def header(self):
        """
        """
        if self._parent is None:
            raise comma.exceptions.CommaOrphanTableException(
                "table not linked to parent structure")

        if self._parent.header is None:
            raise comma.exceptions.CommaNoHeaderException(
                "CSV file does not appear to have a header and "
                "none was provided; key-based access not possible")

        return self._parent.header

    def __getitem__(self, key):
        ##print("CommaTable.__getitem__", hex(id(self)), key, type(key))

        # if getting a specific row: Return underlying CommaRow
        if type(key) is int:
            return super().__getitem__(key)

        # if getting a range of rows: Create a new CommaTable to
        # serve the subset of rows; since the rows will be the
        # same references, any changes will propagate to the whole
        # file (unlike when creating a slice, which usually creates
        # a copy)

        if type(key) is slice:
            return self.clone(newdata=self.data[key])

        # field-slice, i.e. csv_table["street"]
        if type(key) is str:
            return comma.classes.slices.CommaFieldSlice(
                initlist=self,
                parent=self._parent,
                field_name=key)

    def __setitem__(self, key, value):
        ##print("CommaTable.__setitem__", hex(id(self)), key, type(key), "<==", value, type(value))

        # FIXME: figure out how to make slices have headers
        if type(key) is int or type(key) is slice:
            return super().__setitem__(key, value)

        # field-slice, i.e. csv_table["street"]
        if type(key) is str:

            # check size:
            if len(value) != len(self):
                raise comma.exceptions.CommaBatchException(
                    "not right size")

            for i in range(len(self)):
                self[i][key] = value[i]

            return self